@@ -10,28 +10,35 @@ local st = require "util.stanza";
 
 module:add_feature("jabber:iq:version");
 
-local version = "the best operating system ever!";
+local version;
+
+local query = st.stanza("query", {xmlns = "jabber:iq:version"})
+	:tag("name"):text("Prosody"):up()
+	:tag("version"):text(prosody.version):up();
 
 if not module:get_option("hide_os_type") then
 	if os.getenv("WINDIR") then
 		version = "Windows";
 	else
-		local uname = io.popen("uname");
-		if uname then
-			version = uname:read("*a");
-		else
-			version = "an OS";
+		local os_version_command = module:get_option("os_version_command");
+		local ok, pposix = pcall(require, "util.pposix");
+		if not os_version_command and (ok and pposix and pposix.uname) then
+			version = pposix.uname().sysname;
+		end
+		if not version then
+			local uname = io.popen(os_version_command or "uname");
+			if uname then
+				version = uname:read("*a");
+			end
+			uname:close();
 		end
 	end
+	if version then
+		version = version:match("^%s*(.-)%s*$") or version;
+		query:tag("os"):text(version):up();
+	end
 end
 
-version = version:match("^%s*(.-)%s*$") or version;
-
-local query = st.stanza("query", {xmlns = "jabber:iq:version"})
-	:tag("name"):text("Prosody"):up()
-	:tag("version"):text(prosody.version):up()
-	:tag("os"):text(version);
-
 module:hook("iq/host/jabber:iq:version:query", function(event)
 	local stanza = event.stanza;
 	if stanza.attr.type == "get" and stanza.attr.to == module.host then
