@@ -9,7 +9,6 @@
 local select = select;
 local pairs, ipairs = pairs, ipairs;
 
-local datamanager = require "util.datamanager";
 local datetime = require "util.datetime";
 
 local dataform = require "util.dataforms";
@@ -19,14 +18,13 @@ local jid_bare = require "util.jid".bare;
 local jid_prep = require "util.jid".prep;
 local st = require "util.stanza";
 local log = require "util.logger".init("mod_muc");
-local multitable_new = require "util.multitable".new;
 local t_insert, t_remove = table.insert, table.remove;
 local setmetatable = setmetatable;
 local base64 = require "util.encodings".base64;
 local md5 = require "util.hashes".md5;
 
 local muc_domain = nil; --module:get_host();
-local default_history_length = 20;
+local default_history_length, max_history_length = 20, math.huge;
 
 ------------
 local function filter_xmlns_from_array(array, filters)
@@ -90,6 +88,10 @@ local function getText(stanza, path) return getUsingPath(stanza, path, true); en
 local room_mt = {};
 room_mt.__index = room_mt;
 
+function room_mt:__tostring()
+	return "MUC room ("..self.jid..")";
+end
+
 function room_mt:get_default_role(affiliation)
 	if affiliation == "owner" or affiliation == "admin" then
 		return "moderator";
@@ -133,7 +135,6 @@ function room_mt:broadcast_message(stanza, historic)
 		stanza = st.clone(stanza);
 		stanza.attr.to = "";
 		local stamp = datetime.datetime();
-		local chars = #tostring(stanza);
 		stanza:tag("delay", {xmlns = "urn:xmpp:delay", from = muc_domain, stamp = stamp}):up(); -- XEP-0203
 		stanza:tag("x", {xmlns = "jabber:x:delay", from = muc_domain, stamp = datetime.legacy()}):up(); -- XEP-0091 (deprecated)
 		local entry = { stanza = stanza, stamp = stamp };
@@ -185,7 +186,6 @@ function room_mt:send_history(to, stanza)
 
 		local n = 0;
 		local charcount = 0;
-		local stanzacount = 0;
 		
 		for i=#history,1,-1 do
 			local entry = history[i];
@@ -213,6 +213,7 @@ function room_mt:send_history(to, stanza)
 end
 
 function room_mt:get_disco_info(stanza)
+	local count = 0; for _ in pairs(self._occupants) do count = count + 1; end
 	return st.reply(stanza):query("http://jabber.org/protocol/disco#info")
 		:tag("identity", {category="conference", type="text", name=self:get_name()}):up()
 		:tag("feature", {var="http://jabber.org/protocol/muc"}):up()
@@ -224,7 +225,8 @@ function room_mt:get_disco_info(stanza)
 		:tag("feature", {var=self._data.whois ~= "anyone" and "muc_semianonymous" or "muc_nonanonymous"}):up()
 		:add_child(dataform.new({
 			{ name = "FORM_TYPE", type = "hidden", value = "http://jabber.org/protocol/muc#roominfo" },
-			{ name = "muc#roominfo_description", label = "Description"}
+			{ name = "muc#roominfo_description", label = "Description"},
+			{ name = "muc#roominfo_occupants", label = "Number of occupants", value = tostring(count) }
 		}):form({["muc#roominfo_description"] = self:get_description()}, 'result'))
 	;
 end
@@ -337,6 +339,42 @@ end
 function room_mt:get_changesubject()
 	return self._data.changesubject;
 end
+function room_mt:get_historylength()
+	return self._data.history_length or default_history_length;
+end
+function room_mt:set_historylength(length)
+	length = math.min(tonumber(length) or default_history_length, max_history_length or math.huge);
+	if length == default_history_length then
+		length = nil;
+	end
+	self._data.history_length = length;
+end
+
+
+local function construct_stanza_id(room, stanza)
+	local from_jid, to_nick = stanza.attr.from, stanza.attr.to;
+	local from_nick = room._jid_nick[from_jid];
+	local occupant = room._occupants[to_nick];
+	local to_jid = occupant.jid;
+	
+	return from_nick, to_jid, base64.encode(to_jid.."\0"..stanza.attr.id.."\0"..md5(from_jid));
+end
+local function deconstruct_stanza_id(room, stanza)
+	local from_jid_possiblybare, to_nick = stanza.attr.from, stanza.attr.to;
+	local from_jid, id, to_jid_hash = (base64.decode(stanza.attr.id) or ""):match("^(.+)%z(.*)%z(.+)$");
+	local from_nick = room._jid_nick[from_jid];
+
+	if not(from_nick) then return; end
+	if not(from_jid_possiblybare == from_jid or from_jid_possiblybare == jid_bare(from_jid)) then return; end
+
+	local occupant = room._occupants[to_nick];
+	for to_jid in pairs(occupant and occupant.sessions or {}) do
+		if md5(to_jid) == to_jid_hash then
+			return from_nick, to_jid, id;
+		end
+	end
+end
+
 
 function room_mt:handle_to_occupant(origin, stanza) -- PM, vCards, etc
 	local from, to = stanza.attr.from, stanza.attr.to;
@@ -356,6 +394,7 @@ function room_mt:handle_to_occupant(origin, stanza) -- PM, vCards, etc
 		elseif type == "unavailable" then -- unavailable
 			if current_nick then
 				log("debug", "%s leaving %s", current_nick, room);
+				self._jid_nick[from] = nil;
 				local occupant = self._occupants[current_nick];
 				local new_jid = next(occupant.sessions);
 				if new_jid == from then new_jid = next(occupant.sessions, new_jid); end
@@ -380,7 +419,6 @@ function room_mt:handle_to_occupant(origin, stanza) -- PM, vCards, etc
 					self:broadcast_presence(pr, from);
 					self._occupants[current_nick] = nil;
 				end
-				self._jid_nick[from] = nil;
 			end
 		elseif not type then -- available
 			if current_nick then
@@ -488,25 +526,14 @@ function room_mt:handle_to_occupant(origin, stanza) -- PM, vCards, etc
 			end
 		end
 	elseif not current_nick then -- not in room
-		if type == "error" or type == "result" then
-			local id = stanza.name == "iq" and stanza.attr.id and base64.decode(stanza.attr.id);
-			local _nick, _id, _hash = (id or ""):match("^(.+)%z(.*)%z(.+)$");
-			local occupant = self._occupants[stanza.attr.to];
-			if occupant and _nick and self._jid_nick[_nick] and _id and _hash then
-				local id, _to = stanza.attr.id;
-				for jid in pairs(occupant.sessions) do
-					if md5(jid) == _hash then
-						_to = jid;
-						break;
-					end
-				end
-				if _to then
-					stanza.attr.to, stanza.attr.from, stanza.attr.id = _to, self._jid_nick[_nick], _id;
-					self:_route_stanza(stanza);
-					stanza.attr.to, stanza.attr.from, stanza.attr.id = to, from, id;
-				end
+		if (type == "error" or type == "result") and stanza.name == "iq" then
+			local id = stanza.attr.id;
+			stanza.attr.from, stanza.attr.to, stanza.attr.id = deconstruct_stanza_id(self, stanza);
+			if stanza.attr.id then
+				self:_route_stanza(stanza);
 			end
-		else
+			stanza.attr.from, stanza.attr.to, stanza.attr.id = from, to, id;
+		elseif type ~= "error" then
 			origin.send(st.error_reply(stanza, "cancel", "not-acceptable"));
 		end
 	elseif stanza.name == "message" and type == "groupchat" then -- groupchat messages not allowed in PM
@@ -518,16 +545,28 @@ function room_mt:handle_to_occupant(origin, stanza) -- PM, vCards, etc
 		local o_data = self._occupants[to];
 		if o_data then
 			log("debug", "%s sent private stanza to %s (%s)", from, to, o_data.jid);
-			local jid = o_data.jid;
-			local bare = jid_bare(jid);
-			stanza.attr.to, stanza.attr.from = jid, current_nick;
-			local id = stanza.attr.id;
-			if stanza.name=='iq' and type=='get' and stanza.tags[1].attr.xmlns == 'vcard-temp' and bare ~= jid then
-				stanza.attr.to = bare;
-				stanza.attr.id = base64.encode(jid.."\0"..id.."\0"..md5(from));
+			if stanza.name == "iq" then
+				local id = stanza.attr.id;
+				if stanza.attr.type == "get" or stanza.attr.type == "set" then
+					stanza.attr.from, stanza.attr.to, stanza.attr.id = construct_stanza_id(self, stanza);
+				else
+					stanza.attr.from, stanza.attr.to, stanza.attr.id = deconstruct_stanza_id(self, stanza);
+				end
+				if type == 'get' and stanza.tags[1].attr.xmlns == 'vcard-temp' then
+					stanza.attr.to = jid_bare(stanza.attr.to);
+				end
+				if stanza.attr.id then
+					self:_route_stanza(stanza);
+				end
+				stanza.attr.from, stanza.attr.to, stanza.attr.id = from, to, id;
+			else -- message
+				stanza.attr.from = current_nick;
+				for jid in pairs(o_data.sessions) do
+					stanza.attr.to = jid;
+					self:_route_stanza(stanza);
+				end
+				stanza.attr.from, stanza.attr.to = from, to;
 			end
-			self:_route_stanza(stanza);
-			stanza.attr.to, stanza.attr.from, stanza.attr.id = to, from, id;
 		elseif type ~= "error" and type ~= "result" then -- recipient not in room
 			origin.send(st.error_reply(stanza, "cancel", "item-not-found", "Recipient not in room"));
 		end
@@ -541,10 +580,9 @@ function room_mt:send_form(origin, stanza)
 end
 
 function room_mt:get_form_layout()
-	local title = "Configuration for "..self.jid;
-	return dataform.new({
-		title = title,
-		instructions = title,
+	local form = dataform.new({
+		title = "Configuration for "..self.jid,
+		instructions = "Complete and submit this form to configure the room.",
 		{
 			name = 'FORM_TYPE',
 			type = 'hidden',
@@ -606,8 +644,15 @@ function room_mt:get_form_layout()
 			type = 'boolean',
 			label = 'Make Room Members-Only?',
 			value = self:is_members_only()
+		},
+		{
+			name = 'muc#roomconfig_historylength',
+			type = 'text-single',
+			label = 'Maximum Number of History Messages Returned by Room',
+			value = tostring(self:get_historylength())
 		}
 	});
+	return module:fire_event("muc-config-form", { room = self, form = form }) or form;
 end
 
 local valid_whois = {
@@ -628,6 +673,10 @@ function room_mt:process_form(origin, stanza)
 
 	local dirty = false
 
+	local event = { room = self, fields = fields, changed = dirty };
+	module:fire_event("muc-config-submitted", event);
+	dirty = event.changed or dirty;
+
 	local name = fields['muc#roomconfig_roomname'];
 	if name ~= self:get_name() then
 		self:set_name(name);
@@ -657,6 +706,11 @@ function room_mt:process_form(origin, stanza)
 	dirty = dirty or (self:get_changesubject() ~= (not changesubject and true or nil))
 	module:log('debug', 'changesubject=%s', changesubject and "true" or "false")
 
+	local historylength = tonumber(fields['muc#roomconfig_historylength']);
+	dirty = dirty or (historylength and (self:get_historylength() ~= historylength));
+	module:log('debug', 'historylength=%s', historylength)
+
+
 	local whois = fields['muc#roomconfig_whois'];
 	if not valid_whois[whois] then
 	    origin.send(st.error_reply(stanza, 'cancel', 'bad-request', "Invalid value for 'whois'"));
@@ -675,6 +729,7 @@ function room_mt:process_form(origin, stanza)
 	self:set_persistent(persistent);
 	self:set_hidden(not public);
 	self:set_changesubject(changesubject);
+	self:set_historylength(historylength);
 
 	if self.save then self:save(true); end
 	origin.send(st.reply(stanza));
@@ -712,15 +767,16 @@ function room_mt:destroy(newjid, reason, password)
 		self._occupants[nick] = nil;
 	end
 	self:set_persistent(false);
+	module:fire_event("muc-room-destroyed", { room = self });
 end
 
 function room_mt:handle_to_room(origin, stanza) -- presence changes and groupchat messages, along with disco/etc
 	local type = stanza.attr.type;
 	local xmlns = stanza.tags[1] and stanza.tags[1].attr.xmlns;
 	if stanza.name == "iq" then
-		if xmlns == "http://jabber.org/protocol/disco#info" and type == "get" then
+		if xmlns == "http://jabber.org/protocol/disco#info" and type == "get" and not stanza.tags[1].attr.node then
 			origin.send(self:get_disco_info(stanza));
-		elseif xmlns == "http://jabber.org/protocol/disco#items" and type == "get" then
+		elseif xmlns == "http://jabber.org/protocol/disco#items" and type == "get" and not stanza.tags[1].attr.node then
 			origin.send(self:get_disco_items(stanza));
 		elseif xmlns == "http://jabber.org/protocol/muc#admin" then
 			local actor = stanza.attr.from;
@@ -804,7 +860,7 @@ function room_mt:handle_to_room(origin, stanza) -- presence changes and groupcha
 			elseif stanza.attr.type == "set" then
 				local child = stanza.tags[1].tags[1];
 				if not child then
-					origin.send(st.error_reply(stanza, "auth", "bad-request"));
+					origin.send(st.error_reply(stanza, "modify", "bad-request"));
 				elseif child.name == "destroy" then
 					local newjid = child.attr.jid;
 					local reason, password;
@@ -826,13 +882,12 @@ function room_mt:handle_to_room(origin, stanza) -- presence changes and groupcha
 		end
 	elseif stanza.name == "message" and type == "groupchat" then
 		local from, to = stanza.attr.from, stanza.attr.to;
-		local room = jid_bare(to);
 		local current_nick = self._jid_nick[from];
 		local occupant = self._occupants[current_nick];
 		if not occupant then -- not in room
 			origin.send(st.error_reply(stanza, "cancel", "not-acceptable"));
 		elseif occupant.role == "visitor" then
-			origin.send(st.error_reply(stanza, "cancel", "forbidden"));
+			origin.send(st.error_reply(stanza, "auth", "forbidden"));
 		else
 			local from = stanza.attr.from;
 			stanza.attr.from = current_nick;
@@ -843,10 +898,10 @@ function room_mt:handle_to_room(origin, stanza) -- presence changes and groupcha
 					self:set_subject(current_nick, subject); -- TODO use broadcast_message_stanza
 				else
 					stanza.attr.from = from;
-					origin.send(st.error_reply(stanza, "cancel", "forbidden"));
+					origin.send(st.error_reply(stanza, "auth", "forbidden"));
 				end
 			else
-				self:broadcast_message(stanza, true);
+				self:broadcast_message(stanza, self:get_historylength() > 0 and stanza:get_child("body"));
 			end
 			stanza.attr.from = from;
 		end
@@ -864,8 +919,8 @@ function room_mt:handle_to_room(origin, stanza) -- presence changes and groupcha
 		elseif type ~= "error" and type ~= "result" then
 			origin.send(st.error_reply(stanza, "cancel", "service-unavailable"));
 		end
-	elseif stanza.name == "message" and not stanza.attr.type and #stanza.tags == 1 and self._jid_nick[stanza.attr.from]
-		and stanza.tags[1].name == "x" and stanza.tags[1].attr.xmlns == "http://jabber.org/protocol/muc#user" then
+	elseif stanza.name == "message" and not(type == "chat" or type == "error" or type == "groupchat" or type == "headline") and #stanza.tags == 1
+		and self._jid_nick[stanza.attr.from] and stanza.tags[1].name == "x" and stanza.tags[1].attr.xmlns == "http://jabber.org/protocol/muc#user" then
 		local x = stanza.tags[1];
 		local payload = (#x.tags == 1 and x.tags[1]);
 		if payload and payload.name == "invite" and payload.attr.to then
@@ -929,21 +984,23 @@ function room_mt:set_affiliation(actor, jid, affiliation, callback, reason)
 	if affiliation and affiliation ~= "outcast" and affiliation ~= "owner" and affiliation ~= "admin" and affiliation ~= "member" then
 		return nil, "modify", "not-acceptable";
 	end
-	local actor_affiliation = self:get_affiliation(actor);
-	local target_affiliation = self:get_affiliation(jid);
-	if target_affiliation == affiliation then -- no change, shortcut
-		if callback then callback(); end
-		return true;
-	end
-	if actor_affiliation ~= "owner" then
-		if actor_affiliation ~= "admin" or target_affiliation == "owner" or target_affiliation == "admin" then
-			return nil, "cancel", "not-allowed";
+	if actor ~= true then
+		local actor_affiliation = self:get_affiliation(actor);
+		local target_affiliation = self:get_affiliation(jid);
+		if target_affiliation == affiliation then -- no change, shortcut
+			if callback then callback(); end
+			return true;
 		end
-	elseif target_affiliation == "owner" and jid_bare(actor) == jid then -- self change
-		local is_last = true;
-		for j, aff in pairs(self._affiliations) do if j ~= jid and aff == "owner" then is_last = false; break; end end
-		if is_last then
-			return nil, "cancel", "conflict";
+		if actor_affiliation ~= "owner" then
+			if affiliation == "owner" or affiliation == "admin" or actor_affiliation ~= "admin" or target_affiliation == "owner" or target_affiliation == "admin" then
+				return nil, "cancel", "not-allowed";
+			end
+		elseif target_affiliation == "owner" and jid_bare(actor) == jid then -- self change
+			local is_last = true;
+			for j, aff in pairs(self._affiliations) do if j ~= jid and aff == "owner" then is_last = false; break; end end
+			if is_last then
+				return nil, "cancel", "conflict";
+			end
 		end
 	end
 	self._affiliations[jid] = affiliation;
@@ -997,11 +1054,12 @@ function room_mt:get_role(nick)
 	return session and session.role or nil;
 end
 function room_mt:can_set_role(actor_jid, occupant_jid, role)
-	local actor = self._occupants[self._jid_nick[actor_jid]];
 	local occupant = self._occupants[occupant_jid];
-	
-	if not occupant or not actor then return nil, "modify", "not-acceptable"; end
+	if not occupant or not actor_jid then return nil, "modify", "not-acceptable"; end
 
+	if actor_jid == true then return true; end
+
+	local actor = self._occupants[self._jid_nick[actor_jid]];
 	if actor.role == "moderator" then
 		if occupant.affiliation ~= "owner" and occupant.affiliation ~= "admin" then
 			if actor.affiliation == "owner" or actor.affiliation == "admin" then
@@ -1098,10 +1156,17 @@ function _M.new_room(jid, config)
 		_occupants = {};
 		_data = {
 		    whois = 'moderators';
-		    history_length = (config and config.history_length);
+		    history_length = math.min((config and config.history_length)
+		    	or default_history_length, max_history_length);
 		};
 		_affiliations = {};
 	}, room_mt);
 end
 
+function _M.set_max_history_length(_max_history_length)
+	max_history_length = _max_history_length or math.huge;
+end
+
+_M.room_mt = room_mt;
+
 return _M;
