@@ -13,7 +13,7 @@
 * POSIX support functions for Lua
 */
 
-#define MODULE_VERSION "0.3.3"
+#define MODULE_VERSION "0.3.6"
 
 #include <stdlib.h>
 #include <math.h>
@@ -22,6 +22,7 @@
 #include <sys/resource.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/utsname.h>
 #include <fcntl.h>
 
 #include <syslog.h>
@@ -31,8 +32,19 @@
 #include <string.h>
 #include <errno.h>
 #include "lua.h"
+#include "lualib.h"
 #include "lauxlib.h"
 
+#include <fcntl.h>
+#if defined(__linux__) && defined(_GNU_SOURCE)
+#include <linux/falloc.h>
+#endif
+
+#if (defined(_SVID_SOURCE) && !defined(WITHOUT_MALLINFO))
+	#include <malloc.h>
+	#define WITH_MALLINFO
+#endif
+
 /* Daemonization support */
 
 static int lc_daemonize(lua_State *L)
@@ -77,6 +89,10 @@ static int lc_daemonize(lua_State *L)
 	close(0);
 	close(1);
 	close(2);
+	/* Make sure accidental use of FDs 0, 1, 2 don't cause weirdness */
+	open("/dev/null", O_RDONLY);
+	open("/dev/null", O_WRONLY);
+	open("/dev/null", O_WRONLY);
 
 	/* Final fork, use it wisely */
 	if(fork())
@@ -188,12 +204,13 @@ int level_constants[] = 	{
 			};
 int lc_syslog_log(lua_State* L)
 {
-	int level = luaL_checkoption(L, 1, "notice", level_strings);
-	level = level_constants[level];
+	int level = level_constants[luaL_checkoption(L, 1, "notice", level_strings)];
 
-	luaL_checkstring(L, 2);
+	if(lua_gettop(L) == 3)
+		syslog(level, "%s: %s", luaL_checkstring(L, 2), luaL_checkstring(L, 3));
+	else
+		syslog(level, "%s", lua_tostring(L, 2));
 
-	syslog(level, "%s", lua_tostring(L, 2));
 	return 0;
 }
 
@@ -359,6 +376,66 @@ int lc_setgid(lua_State* L)
 	return 2;
 }
 
+int lc_initgroups(lua_State* L)
+{
+	int ret;
+	gid_t gid;
+	struct passwd *p;
+
+	if(!lua_isstring(L, 1))
+	{
+		lua_pushnil(L);
+		lua_pushstring(L, "invalid-username");
+		return 2;
+	}
+	p = getpwnam(lua_tostring(L, 1));
+	if(!p)
+	{
+		lua_pushnil(L);
+		lua_pushstring(L, "no-such-user");
+		return 2;
+	}
+	if(lua_gettop(L) < 2)
+		lua_pushnil(L);
+	switch(lua_type(L, 2))
+	{
+	case LUA_TNIL:
+		gid = p->pw_gid;
+		break;
+	case LUA_TNUMBER:
+		gid = lua_tointeger(L, 2);
+		break;
+	default:
+		lua_pushnil(L);
+		lua_pushstring(L, "invalid-gid");
+		return 2;
+	}
+	ret = initgroups(lua_tostring(L, 1), gid);
+	if(ret)
+	{
+		switch(errno)
+		{
+		case ENOMEM:
+			lua_pushnil(L);
+			lua_pushstring(L, "no-memory");
+			break;
+		case EPERM:
+			lua_pushnil(L);
+			lua_pushstring(L, "permission-denied");
+			break;
+		default:
+			lua_pushnil(L);
+			lua_pushstring(L, "unknown-error");
+		}
+	}
+	else
+	{
+		lua_pushboolean(L, 1);
+		lua_pushnil(L);
+	}
+	return 2;
+}
+
 int lc_umask(lua_State* L)
 {
 	char old_mode_string[7];
@@ -407,6 +484,9 @@ int string2resource(const char *s) {
 	if (!strcmp(s, "MEMLOCK")) return RLIMIT_MEMLOCK;
 	if (!strcmp(s, "NPROC")) return RLIMIT_NPROC;
 	if (!strcmp(s, "RSS")) return RLIMIT_RSS;
+#endif
+#ifdef RLIMIT_NICE
+	if (!strcmp(s, "NICE")) return RLIMIT_NICE;
 #endif
 	return -1;
 }
@@ -420,6 +500,7 @@ int lc_setrlimit(lua_State *L) {
 	if(arguments < 1 || arguments > 3) {
 		lua_pushboolean(L, 0);
 		lua_pushstring(L, "incorrect-arguments");
+		return 2;
 	}
 
 	resource = luaL_checkstring(L, 1);
@@ -497,6 +578,134 @@ int lc_abort(lua_State* L)
 	return 0;
 }
 
+int lc_uname(lua_State* L)
+{
+	struct utsname uname_info;
+	if(uname(&uname_info) != 0)
+	{
+		lua_pushnil(L);
+		lua_pushstring(L, strerror(errno));
+		return 2;
+	}
+	lua_newtable(L);
+	lua_pushstring(L, uname_info.sysname);
+	lua_setfield(L, -2, "sysname");
+	lua_pushstring(L, uname_info.nodename);
+	lua_setfield(L, -2, "nodename");
+	lua_pushstring(L, uname_info.release);
+	lua_setfield(L, -2, "release");
+	lua_pushstring(L, uname_info.version);
+	lua_setfield(L, -2, "version");
+	lua_pushstring(L, uname_info.machine);
+	lua_setfield(L, -2, "machine");
+	return 1;
+}
+
+int lc_setenv(lua_State* L)
+{
+	const char *var = luaL_checkstring(L, 1);
+	const char *value;
+
+	/* If the second argument is nil or nothing, unset the var */
+	if(lua_isnoneornil(L, 2))
+	{
+		if(unsetenv(var) != 0)
+		{
+			lua_pushnil(L);
+			lua_pushstring(L, strerror(errno));
+			return 2;
+		}
+		lua_pushboolean(L, 1);
+		return 1;
+	}
+
+	value = luaL_checkstring(L, 2);
+
+	if(setenv(var, value, 1) != 0)
+	{
+		lua_pushnil(L);
+		lua_pushstring(L, strerror(errno));
+		return 2;
+	}
+
+	lua_pushboolean(L, 1);
+	return 1;
+}
+
+#ifdef WITH_MALLINFO
+int lc_meminfo(lua_State* L)
+{
+	struct mallinfo info = mallinfo();
+	lua_newtable(L);
+	/* This is the total size of memory allocated with sbrk by malloc, in bytes. */
+	lua_pushinteger(L, info.arena);
+	lua_setfield(L, -2, "allocated");
+	/* This is the total size of memory allocated with mmap, in bytes. */
+	lua_pushinteger(L, info.hblkhd);
+	lua_setfield(L, -2, "allocated_mmap");
+	/* This is the total size of memory occupied by chunks handed out by malloc. */
+	lua_pushinteger(L, info.uordblks);
+	lua_setfield(L, -2, "used");
+	/* This is the total size of memory occupied by free (not in use) chunks. */
+	lua_pushinteger(L, info.fordblks);
+	lua_setfield(L, -2, "unused");
+	/* This is the size of the top-most releasable chunk that normally borders the
+	   end of the heap (i.e., the high end of the virtual address space's data segment). */
+	lua_pushinteger(L, info.keepcost);
+	lua_setfield(L, -2, "returnable");
+	return 1;
+}
+#endif
+
+/* File handle extraction blatantly stolen from
+ * https://github.com/rrthomas/luaposix/blob/master/lposix.c#L631
+ * */
+
+#if _XOPEN_SOURCE >= 600 || _POSIX_C_SOURCE >= 200112L || defined(_GNU_SOURCE)
+int lc_fallocate(lua_State* L)
+{
+	off_t offset, len;
+	FILE *f = *(FILE**) luaL_checkudata(L, 1, LUA_FILEHANDLE);
+
+	offset = luaL_checkinteger(L, 2);
+	len = luaL_checkinteger(L, 3);
+
+#if defined(__linux__) && defined(_GNU_SOURCE)
+	if(fallocate(fileno(f), FALLOC_FL_KEEP_SIZE, offset, len) == 0)
+	{
+		lua_pushboolean(L, 1);
+		return 1;
+	}
+
+	if(errno != ENOSYS && errno != EOPNOTSUPP)
+	{
+		lua_pushnil(L);
+		lua_pushstring(L, strerror(errno));
+		return 2;
+	}
+#else
+#warning Only using posix_fallocate() fallback.
+#warning Linux fallocate() is strongly recommended if available: recompile with -D_GNU_SOURCE
+#warning Note that posix_fallocate() will still be used on filesystems that dont support fallocate()
+#endif
+
+	if(posix_fallocate(fileno(f), offset, len) == 0)
+	{
+		lua_pushboolean(L, 1);
+		return 1;
+	}
+	else
+	{
+		lua_pushnil(L);
+		lua_pushstring(L, strerror(errno));
+		/* posix_fallocate() can leave a bunch of NULs at the end, so we cut that
+		 * this assumes that offset == length of the file */
+		ftruncate(fileno(f), offset);
+		return 2;
+	}
+}
+#endif
+
 /* Register functions */
 
 int luaopen_util_pposix(lua_State *L)
@@ -517,6 +726,7 @@ int luaopen_util_pposix(lua_State *L)
 
 		{ "setuid", lc_setuid },
 		{ "setgid", lc_setgid },
+		{ "initgroups", lc_initgroups },
 
 		{ "umask", lc_umask },
 
@@ -525,6 +735,18 @@ int luaopen_util_pposix(lua_State *L)
 		{ "setrlimit", lc_setrlimit },
 		{ "getrlimit", lc_getrlimit },
 
+		{ "uname", lc_uname },
+
+		{ "setenv", lc_setenv },
+
+#ifdef WITH_MALLINFO
+		{ "meminfo", lc_meminfo },
+#endif
+
+#if _XOPEN_SOURCE >= 600 || _POSIX_C_SOURCE >= 200112L || defined(_GNU_SOURCE)
+		{ "fallocate", lc_fallocate },
+#endif
+
 		{ NULL, NULL }
 	};
 
@@ -537,4 +759,4 @@ int luaopen_util_pposix(lua_State *L)
 	lua_setfield(L, -2, "_VERSION");
 
 	return 1;
-};
+}
