@@ -8,8 +8,10 @@
 
 --[[ Iterators ]]--
 
+local it = {};
+
 -- Reverse an iterator
-function reverse(f, s, var)
+function it.reverse(f, s, var)
 	local results = {};
 
 	-- First call the normal iterator
@@ -34,12 +36,12 @@ end
 local function _keys_it(t, key)
 	return (next(t, key));
 end
-function keys(t)
+function it.keys(t)
 	return _keys_it, t;
 end
 
 -- Iterate only over values in a table
-function values(t)
+function it.values(t)
 	local key, val;
 	return function (t)
 		key, val = next(t, key);
@@ -48,7 +50,7 @@ function values(t)
 end
 
 -- Given an iterator, iterate only over unique items
-function unique(f, s, var)
+function it.unique(f, s, var)
 	local set = {};
 	
 	return function ()
@@ -65,7 +67,7 @@ function unique(f, s, var)
 end
 
 --[[ Return the number of items an iterator returns ]]--
-function count(f, s, var)
+function it.count(f, s, var)
 	local x = 0;
 	
 	while true do
@@ -73,13 +75,13 @@ function count(f, s, var)
 		var = ret[1];
 	        if var == nil then break; end
 		x = x + 1;
-	end	
+	end
 	
 	return x;
 end
 
 -- Return the first n items an iterator returns
-function head(n, f, s, var)
+function it.head(n, f, s, var)
 	local c = 0;
 	return function (s, var)
 		if c >= n then
@@ -90,7 +92,16 @@ function head(n, f, s, var)
 	end, s;
 end
 
-function tail(n, f, s, var)
+-- Skip the first n items an iterator returns
+function it.skip(n, f, s, var)
+	for i=1,n do
+		var = f(s, var);
+	end
+	return f, s, var;
+end
+
+-- Return the last n items an iterator returns
+function it.tail(n, f, s, var)
 	local results, count = {}, 0;
 	while true do
 		local ret = { f(s, var) };
@@ -111,8 +122,19 @@ function tail(n, f, s, var)
 	--return reverse(head(n, reverse(f, s, var)));
 end
 
+local function _ripairs_iter(t, key) if key > 1 then return key-1, t[key-1]; end end
+function it.ripairs(t)
+	return _ripairs_iter, t, #t+1;
+end
+
+local function _range_iter(max, curr) if curr < max then return curr + 1; end end
+function it.range(x, y)
+	if not y then x, y = 1, x; end -- Default to 1..x if y not given
+	return _range_iter, y, x-1;
+end
+
 -- Convert the values returned by an iterator to an array
-function it2array(f, s, var)
+function it.to_array(f, s, var)
 	local t, var = {};
 	while true do
 		var = f(s, var);
@@ -122,10 +144,10 @@ function it2array(f, s, var)
 	return t;
 end
 
--- Treat the return of an iterator as key,value pairs, 
+-- Treat the return of an iterator as key,value pairs,
 -- and build a table
-function it2table(f, s, var)
-	local t, var = {};
+function it.to_table(f, s, var)
+	local t, var2 = {};
 	while true do
 		var, var2 = f(s, var);
 	        if var == nil then break; end
@@ -133,3 +155,5 @@ function it2table(f, s, var)
 	end
 	return t;
 end
+
+return it;
